#
# Copyright (c) 2007 Patrick Van Oosterwijck
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

"""Balloon escape game engine"""

import sys
import random
import pygame
import gameobj
import objects
import world


# Menu colors

bgcolor = 77, 100, 118
fgcolor = 255, 255, 255
skyblue = 192, 230, 237
fuelcolor = 195, 18, 18
liftcolor = 253, 186, 68
dropcolor = 89, 131, 248


class GameEngine:
    """This is the central game engine class.  All objects are managed here and their
    behavior is tied together to create the game"""
    
    def __init__(self):
        """Init function when creating the game engine object"""
        # Menu definition
        self.menu_list = [('Dare me', 5),
                          ('Tough guy', 3),
                          ('Average Joe', 1),
                          ('Wimpy', 0),
                          ('Chicken', -1)]
        self.menu_sel = 2
        # Game status
        self.game_status_list = \
                [None, "Play!", "You crashed!", "You made it!"]
        self.game_status = 1
        # Create viewport with default size
        self.SetViewportSize((world.trench_width, world.trench_width*0.9))
        # Create the balloon
        self.balloon = objects.Balloon()
        # Create the crow group and set default target number of crows
        self.crow_group = gameobj.RenderGameObj()
        self.crow_count_target = 15
        # Create the parachute group and set default target number
        self.parachute_group = gameobj.RenderGameObj()
        self.parachute_count_target = 15
        # Create the trench object
        self.trench = world.Trench()
        # Create sprite groups
        self.enemy_group = pygame.sprite.Group()
        self.resize_group = gameobj.RenderGameObj()
        self.physics_group = pygame.sprite.Group()
        self.visible_group = gameobj.RenderGameObj()
        self.aniobj_group = gameobj.RenderGameObj()

    def SetViewportSize(self, size):
        """Set the game viewport size"""
        # Save new viewport size
        world.view_width = int(size[0])
        world.view_height = int(size[1])
        # Don't allow the viewport to be more than twice as high as it is wide
        if world.view_height > 2*world.view_width:
            world.view_height = 2*world.view_width
        # Create screen with view size
        self.screen = pygame.display.set_mode(
                (world.view_width, world.view_height))
        # Create and set background
        self.background = self.screen.copy()
        self.background.fill(skyblue)
        self.screen.blit(self.background, self.background.get_rect())
        pygame.display.update()
        # Calculate the screen viewport scale
        world.scr_scale = float(world.view_width) / world.trench_width
        try:
            # Reload textures and recalculate coordinates based on scale
            self.trench.Load(scrscale=world.scr_scale)
            self.resize_group.resized_screen(world.scr_scale)
            self.CenterBalloon()
        except AttributeError:
            pass
        # Create menu box: definitions
        menucover = 0.6
        menurect = pygame.Rect(0, 0, world.view_width * menucover,
                                     world.view_height * menucover)
        # Create menu surface with frame
        self.menu_surface = pygame.Surface(menurect.size)
        pygame.draw.rect(self.menu_surface, bgcolor, menurect)
        pygame.draw.rect(self.menu_surface, fgcolor,
                         menurect.inflate(-5, -5), 2)
        # Render game menu messages
        font = pygame.font.Font(None, int(50*world.scr_scale))
        self.game_menu_rects = []
        for i in range(len(self.menu_list)):
            textsurface = font.render(self.menu_list[i][0], True, fgcolor)
            textrect = textsurface.get_rect()
            textrect.move_ip(menurect.left+(menurect.width-textrect.width)/2,
                          menurect.top+(textrect.height*(i*1.7+3.8)))
            self.game_menu_rects.append(textrect)
            self.menu_surface.blit(textsurface, textrect)
        # Move menu rectangle and save it
        menurect.move_ip(world.view_width * (1.0-menucover)/2,
                      world.view_height * (1.0-menucover)/2)
        self.menu_rect = menurect
        # Render game status messages
        font = pygame.font.Font(None, int(80*world.scr_scale))
        self.game_status_gfx = []
        for i in range(len(self.game_status_list)):
            textsurface = font.render(self.game_status_list[i], True, fgcolor)
            textrect = textsurface.get_rect()
            textrect.move_ip(menurect.left+(menurect.width-textrect.width)/2,
                          menurect.top+textrect.height*0.8)
            self.game_status_gfx.append((textsurface, textrect))
        # Convert game menu rectangles to screen coordinates for select box
        for r in self.game_menu_rects:
            r.move_ip(menurect.left, menurect.top)
            r.inflate_ip(10, 10)

    def SetViewportPosition(self, x, y):
        """Set the game viewport position"""
        maxviewy = world.trench_depth - world.view_height/world.scr_scale
        if y > maxviewy: y = maxviewy
        world.view_y = y
        self.visible_group.set_viewport(x, y)

    def CenterBalloon(self):
        """Set the viewport so the balloon is centered"""
        y = self.balloon.position[1] - \
                (world.view_height/world.scr_scale-self.balloon.rect.height)/2
        self.SetViewportPosition(0, y)
        
    def CreateWorld(self, difficulty):
        """Create a new world: generate a new trench with a depth and
        jaggedness based on difficulty, number of crows and refuel
        parachutes based on difficulty, and windyness based on difficulty.
        Difficulty is a number between 0.0 and 10.0."""
        # Set parameters
        world.trench_depth = int(2000 + 400*difficulty)
        self.crow_count_target = int(10 + 1.5*difficulty)
        self.parachute_count_target = int(18 - 1.2*difficulty)
        # No crows or parachutes to start
        self.crow_group.empty()
        self.parachute_group.empty()
        # Clear all groups to release any old objects
        self.enemy_group.empty()
        self.resize_group.empty()
        self.physics_group.empty()
        self.visible_group.empty()
        self.aniobj_group.empty()
        # Add the balloon to the resize, physics and visible groups
        self.balloon.add(self.resize_group, self.physics_group,
                         self.visible_group)
        # Generate a new trench
        self.trench.Load(scrscale=world.scr_scale)
        self.trench.Create(world.trench_width, world.trench_depth,
                           0.1 + 0.15*difficulty, 5.0+0.5*difficulty)
        world.trench = self.trench
        # Add the trench blocks to the enemy, visible and resize groups
        self.enemy_group.add(self.trench.block_group)
        self.visible_group.add(self.trench.block_group)
        self.resize_group.add(self.trench.block_group)
        # Add the trench flags to the visible, resize and aniobj groups
        self.visible_group.add(self.trench.flag_group)
        self.resize_group.add(self.trench.flag_group)
        self.aniobj_group.add(self.trench.flag_group)
        # Add the water to the visible, resize and aniobj groups
        self.visible_group.add(self.trench.water)
        self.resize_group.add(self.trench.water)
        self.aniobj_group.add(self.trench.water)
        # Apply scale to new objects
        self.resize_group.resized_screen(world.scr_scale)
        # Reset the balloon and position it at the bottom of the trench
        self.balloon.Reset()
        self.balloon.position = ((world.trench_width-self.balloon.rect.width)/2,
                        world.trench_depth-1.1*self.balloon.rect.height)
        # Set the viewport to the initial view
        self.CenterBalloon()

    def SpawnCrows(self):
        """Span crows if we are below our target number"""
        # Determine if we should spawn a crow
        if random.uniform(0.0, self.crow_count_target) > \
                    float(len(self.crow_group)):
            # Spawn a new crow
            objects.Crow(self.enemy_group, self.resize_group,
                         self.physics_group, self.visible_group,
                         self.aniobj_group, self.crow_group)

    def SpawnParachutes(self):
        """Span parachutes if we are below our target number"""
        # Determine if we should spawn a parachute
        if random.uniform(0.0, self.parachute_count_target) > \
                    float(len(self.parachute_group)):
            # Spawn a new parachute
            objects.Parachute(self.resize_group,
                         self.physics_group, self.visible_group,
                         self.aniobj_group, self.parachute_group)
        
    def RunFrame(self, frametime):
        """Run one game engine cycle"""
        # Run physics on all relevant objects
        self.physics_group.update()
        # Did any enemies collide with the balloon?
        if gameobj.spritecollide(self.balloon, self.enemy_group, False):
            # Then pop the balloon
            self.balloon.SetDamaged()
        # Did any parachutes collide with the balloon?
        for i in gameobj.spritecollide(
                self.balloon, self.parachute_group, True):
            # Add fuel for each balloon
            self.balloon.fuel_level += random.uniform(40, 60)
            self.balloon.fuel_level = min(self.balloon.fuel_level, 100)
        # Spawn crows and parachutes
        self.SpawnCrows()
        self.SpawnParachutes()
        # Set the viewport to center the balloon
        self.CenterBalloon()
        # Clear all visible objects
        self.visible_group.clear(self.screen, self.background)
        # Draw all groups in back-to-front order, track updated rectangles
        updaterects = []
        updaterects.extend(self.visible_group.draw(self.screen, frametime))
        #updaterects.extend(self.aniobj_group.draw(self.screen, frametime))
        #updaterects.append(self.balloon.Draw(self.screen, frametime))
        #updaterects.extend(self.trench.block_group.draw(self.screen, frametime))
        # Draw HUD
        updaterects.append(self.DrawHUD())
        # Draw menu is not playing
        if self.game_status:
            updaterects.append(self.DrawMenu())
        # Update display
        pygame.display.update(updaterects)
        # Game status check if playing
        if not self.game_status:
            # Crash check
            if self.balloon.position[1] > world.trench_depth:
                self.game_status = 2
            # Win check
            if self.balloon.position[1]+self.balloon.rect.height < 0.0:
                self.game_status = 3

    def DrawMenu(self):
        """Draw the menu, return affected rectangle"""
        # Make sure to keep the balloon static when the menu shows
        self.balloon.SetStatic()
        # Draw menu
        updrect = self.screen.blit(self.menu_surface, self.menu_rect)
        # Draw selection box
        pygame.draw.rect(self.screen, fgcolor,
                         self.game_menu_rects[self.menu_sel], 2)
        # Draw status message
        self.screen.blit(self.game_status_gfx[self.game_status][0],
                         self.game_status_gfx[self.game_status][1])
        return updrect

    def DrawHUD(self):
        """Draw heads up display showing fuel level and lift,
        return affected rectangle"""
        # Not scalable at this point but that doesn't matter since
        # scaling is disabled for now
        # HUD definitions
        hudheight = 200
        hudmargin = 10
        hudwidth = 40
        barmargin = 4
        hudtop = world.view_height - hudheight - hudmargin
        # Draw HUD background rectangle
        hudrect = pygame.Rect(hudmargin, hudtop, hudwidth, hudheight)
        pygame.draw.rect(self.screen, bgcolor, hudrect)
        # Draw the fuel bar
        fuelrect = pygame.Rect(0, 0, 0, 0)
        fuelrect.width = hudwidth/2 - 1.5*barmargin
        fuelrect.height = (hudheight-2*barmargin)*\
                          (self.balloon.fuel_level/100.0)
        fuelrect.left = hudmargin + barmargin
        fuelrect.top = hudtop+hudheight-barmargin-fuelrect.height
        pygame.draw.rect(self.screen, fuelcolor, fuelrect)
        # Draw the lift bar
        liftrect = pygame.Rect(0, 0, 0, 0)
        liftrect.width = hudwidth/2 - 1.5*barmargin
        liftrect.left = hudmargin + hudwidth/2 + 0.5*barmargin
        liftrect.height = int((self.balloon.lift-10.0)*2000)
        if liftrect.height < 0:
            liftrect.height = min(abs(liftrect.height), hudheight/2-barmargin)
            liftrect.top = hudtop + hudheight/2
            pygame.draw.rect(self.screen, dropcolor, liftrect)
        else:
            liftrect.height = min(liftrect.height, hudheight/2-barmargin)
            liftrect.top = hudtop + hudheight/2 - liftrect.height
            pygame.draw.rect(self.screen, liftcolor, liftrect)
        return hudrect
        
    def ProcessEvent(self, event):
        """Process events"""
        if event.type == pygame.QUIT:
            sys.exit()
        if event.type == pygame.VIDEORESIZE:
            self.SetViewportSize(event.size)
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_UP:
                if not self.game_status:
                    self.balloon.SetFlameOn(True)
                else:
                    self.menu_sel = max(0, self.menu_sel-1)
            if event.key == pygame.K_DOWN:
                if not self.game_status:
                    self.balloon.SetVentOpen(True)
                else:
                    self.menu_sel = min(len(self.menu_list)-1, self.menu_sel+1)
            if event.key == pygame.K_ESCAPE:
                if not self.game_status:
                    self.balloon.SetDamaged()
                else:
                    sys.exit()
            if event.key == pygame.K_RETURN and self.game_status:
                self.game_status = 0
                difficulty = self.menu_list[self.menu_sel][1]
                if difficulty < 0:
                    sys.exit()
                else:
                    self.CreateWorld(difficulty)
                    self.RunFrame(0)
                    self.screen.blit(self.background,
                                     self.background.get_rect())
                    pygame.display.update()
        if event.type == pygame.KEYUP and not self.game_status:
            if event.key == pygame.K_UP:
                self.balloon.SetFlameOn(False)
            if event.key == pygame.K_DOWN:
                self.balloon.SetVentOpen(False)

