#
# Copyright (c) 2007 Patrick Van Oosterwijck
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

"""Base classes for game objects"""

import data
import pygame


class GameObj(pygame.sprite.Sprite):
    """
    GameObj class which is derived from the pygame Sprite class.
    It handles all object animation and depends on the data
    structure as defined in PyWeek's Skellington.
    """
    
    def __init__(self, datadir=None, objscale=1.0, scrscale=1.0, *groups):
        """Init function when creating the object"""
        # Invoke base class init first
        pygame.sprite.Sprite.__init__(self, *groups)
        # Save scales
        self.obj_scale = objscale
        self.scr_scale = scrscale
        # If a data directory is specified, load it here
        if datadir: self.Load(datadir, objscale, scrscale)
        # Initialize surface position variables
        self.surface_x, self.surface_y = 0, 0
        # Initialize animation state variables
        self.cur_ani_state = ''
        self.cur_ani_counter = 0
        # Make sure rectangle attributes are available
        self.rect = None
        self.collide_rect = None

    def Load(self, datadir=None, objscale=None, scrscale=None):
        """Load the specified object using the specified scale"""
        # Save the data directory and scales
        if datadir: self.data_dir = datadir
        if objscale: self.obj_scale = objscale
        if scrscale: self.scr_scale = scrscale
        # Calculate the current scale, object and screen combined
        scale = self.obj_scale * self.scr_scale
        # Use Skellington load to read the animation definition
        anidef = data.load(self.data_dir + "/ani.txt").readlines()
        # Define variable for current animation state and
        # animation state dictionary
        curanistate = None
        self.ani_state = {}
        # Create dictionary that will hold images
        self.image = {}
        # Parse all animation definition lines
        for line in anidef:
            # Skip the line if it is a comment
            if line[0] == '#': continue
            # Split the line
            line = line.split()
            # Skip the line if it is empty
            if not len(line): continue
            # Is this a state definition line?
            if line[0].lower() == 'state':
                # Create a new dictionary for the current animation state
                # and add it to the state dictionary
                curanistate = {}
                self.ani_state[line[1]] = curanistate
                # Create a frame list for this animation state
                curanistate['framelist'] = []
                # Create a loop variable, false by default
                curanistate['loop'] = False
                # Create a time variable for the total animation state time
                curanistate['time'] = 0
                # We're done with this line
                continue
            # All other definitions need to have a current state
            # Ignore the line if there is none
            if not curanistate: continue
            # Is this a frame definition line?
            if line[0].lower() == 'frame':
                # Save the frame name
                framename = line[1]
                # Add a frame to the frame list
                curanistate['framelist'].append({'image': framename})
                # Get a reference to this newly added frame
                curanistateframe = curanistate['framelist'][-1]
                # Set default frame parameters
                curanistateframe['time'] = 100
                curanistateframe['xoffs'] = 0.0
                curanistateframe['yoffs'] = 0.0
                # Create image parameter dictionary
                imageparameter = {'name': framename,
                                  'surface': None,
                                  'angle': 0.0,
                                  'scale': 1.0,
                                  'xflip': False,
                                  'yflip': False}
                # Read frame parameters
                i = 2
                while i < len(line):
                    # Is it a time parameter?
                    if line[i] == 'time':
                        # Save the frame time
                        curanistateframe['time'] = int(line[i+1])
                        # On to the next parameter
                        i += 2
                        continue
                    # Is it an X offset parameter?
                    if line[i] == 'xoffs':
                        # Save the frame time
                        curanistateframe['xoffs'] = scale*float(line[i+1])
                        # On to the next parameter
                        i += 2
                        continue
                    # Is it a Y offset parameter?
                    if line[i] == 'yoffs':
                        # Save the frame time
                        curanistateframe['yoffs'] = scale*float(line[i+1])
                        # On to the next parameter
                        i += 2
                        continue
                    # Is it a scale parameter?
                    if line[i] == 'scale':
                        # Set image parameter
                        imageparameter['scale'] = float(line[i+1])
                        # Change the frame name
                        framename += "*%f" % imageparameter['scale']
                        curanistateframe['image'] = framename
                        # On to the next parameter
                        i += 2
                        continue
                    # Is it an angle parameter?
                    if line[i] == 'angle':
                        # Set image parameter
                        imageparameter['angle'] = float(line[i+1])
                        # Change the frame name
                        framename += "/%d" % imageparameter['angle']
                        curanistateframe['image'] = framename
                        # On to the next parameter
                        i += 2
                        continue
                    # Is it an x flip parameter?
                    if line[i] == 'xflip':
                        # Set image parameter
                        imageparameter['xflip'] = True
                        # Change the frame name
                        framename += "+xflip"
                        curanistateframe['image'] = framename
                        # On to the next parameter
                        i += 1
                        continue
                    # Is it a y flip parameter?
                    if line[i] == 'yflip':
                        # Set image parameter
                        imageparameter['yflip'] = True
                        # Change the frame name
                        framename += "+yflip"
                        curanistateframe['image'] = framename
                        # On to the next parameter
                        i += 1
                        continue
                # Add the time of this frame to the animation state time
                curanistate['time'] += curanistateframe['time']
                # Add the frame image to the image dictionary
                self.image[framename] = imageparameter
            # Is this a loop definition line?
            if line[0].lower() == 'loop':
                # Set loop variable to true
                curanistate['loop'] = True
                # We're done with this animation state
                curanistate = None
        # Go through the list of images referenced in the animation
        # definition
        for imageref, image in self.image.items():
            # Use Skellington load to load the image file
            imagefile = data.load(self.data_dir + "/" + image['name'])
            image['surface'] = pygame.image.load(imagefile).convert_alpha()
            # Rotate and scale
            image['surface'] = pygame.transform.rotozoom(image['surface'],
                            image['angle'], scale*image['scale'])
            # Flip image
            image['surface'] = pygame.transform.flip(image['surface'],
                            image['xflip'], image['yflip'])

    def SetViewport(self, x, y):
        """Set the viewport origin to the specified world coordinates"""
        self.surface_x = x
        self.surface_y = y

    def SetState(self, state):
        """Set the animation state to use"""
        self.cur_ani_state = state
        self.cur_ani_counter = 0

    def Draw(self, surface, frametime):
        """Advance animation based on frametime, draw the resulting
        object image to the specified surface and return the bounding
        rectangle. The coordinates where the object should be drawn
        are scaled by the specified scaling factor."""
        # Get shortcut to the current animation state, the animation
        # counter and the screen scale
        cas = self.ani_state[self.cur_ani_state]
        cac = self.cur_ani_counter
        scale = self.scr_scale
        # Add frame time to animation counter
        cac += frametime
        # Limit or loop the counter, depending on loop property
        if cas['loop']:
            cac %= cas['time']
        else:
            cac = min(cac, cas['time'])
        # Save back to animation counter
        self.cur_ani_counter = cac
        # Find the right frame we're in
        ftc = 0
        for frame in cas['framelist']:
            ftc += frame['time']
            if ftc > cac: break
        # Get the image surface for the frame
        ims = self.image[frame['image']]['surface']
        # Calculate the rectangle for the object in the destination surface
        dx = (self.position[0]+frame['xoffs']-self.surface_x)*scale
        dy = (self.position[1]+frame['yoffs']-self.surface_y)*scale
        drect = ims.get_rect().move(dx, dy)
        # Blit the image surface to the destination and return the
        # affected rectangle
        return surface.blit(ims, drect)

    def __str__(self):
        """Show object info when doing a print"""
        return ("Animation states:\n" + str(self.ani_state) + 
                "\nImages loaded:\n" + str(self.image))


class StaticObj(pygame.sprite.Sprite):
    """
    StaticObj class which is derived from the pygame Sprite class.
    It provides compatibility with GameObj for static sprites so it can
    take care of screen scaling and viewport surface adjustments.
    """
    
    def __init__(self, scrscale=1.0, *groups):
        """Init function when creating the object"""
        # Invoke base class init first
        pygame.sprite.Sprite.__init__(self, *groups)
        # Save scale
        self.scr_scale = scrscale
        # Initialize surface position variables
        self.surface_x, self.surface_y = 0, 0
        # Rectangle and image reference
        self.rect = None
        self.collide_rect = None
        self.image_dict = None
        self.image_name = None
        # Miscellaneous value
        self.misc = None

    def Load(self, scrscale=None):
        """Load the specified object using the specified scale.
        For static objects this just applies the scale."""
        if scrscale: self.scr_scale = scrscale

    def SetViewport(self, x, y):
        """Set the viewport origin to the specified world coordinates"""
        self.surface_x = -x
        self.surface_y = -y

    def Draw(self, surface, frametime):
        """Draw the object image to the specified surface and return
        the bounding rectangle. The coordinates where the object should
        be drawn are scaled by the specified scaling factor."""
        # Get scale shortcut
        scale = self.scr_scale
        # Calculate the rectangle for the object in the destination surface
        drect = pygame.Rect(self.rect.left*scale, self.rect.top*scale,
                        self.rect.width, self.rect.height)
        drect.move_ip(self.surface_x*scale, self.surface_y*scale)
        # Blit the image surface to the destination and return the
        # affected rectangle
        return surface.blit(self.image_dict[self.image_name], drect)
        

class RenderGameObj(pygame.sprite.Group):
    """A sprite group that renders game objects efficiently and runs
       through the game object animations based on the frame time.
       This group supports drawing to the screen, but its draw method
       also returns a list of the Rects updated by the draw (and any
       clears in between the last draw and the current one). You
       can use:
       pygame.display.update(rendergameobj_group.draw(screen, frametime, scale))
       to minimize the updated part of the screen."""
    
    def draw(self, surface, frametime):
        spritedict = self.spritedict
        surface_blit = surface.blit
        dirty = self.lostsprites
        self.lostsprites = []
        dirty_append = dirty.append
        for s in self.sprites():
            r = spritedict[s]
            newrect = s.Draw(surface, frametime)
            if r is 0:
                dirty_append(newrect)
            else:
                if newrect.colliderect(r):
                    dirty_append(newrect.union(r))
                else:
                    dirty_append(newrect)
                    dirty_append(r)
            spritedict[s] = newrect
        return dirty

    def resized_screen(self, scale):
        for s in self.sprites():
            s.Load(scrscale=scale)

    def set_viewport(self, x, y):
        for s in self.sprites():
            s.SetViewport(x, y)


def spritecollide(sprite, group, dokill):
    """gameobj.spritecollide(sprite, group, dokill) -> list
       collision detection between sprite and group

       Given a sprite and a group of sprites, this will
       return a list of all the sprites that intersect
       the given sprite.
       All sprites must have a "collide_rect" value, which is
       a rectangle of the collision area.  If the dokill
       argument is True, the sprites that do collide will be
       automatically removed from all groups.
       This is my own version of the function that separates
       the visual rectangle "rect" from the collission
       rectangle "collide_rect"
       """
    crashed = []
    if sprite.collide_rect:
        spritecollide = sprite.collide_rect.colliderect
    else:
        spritecollide = sprite.rect.colliderect
    if dokill:
        for s in group.sprites():
            if s.collide_rect:
                if spritecollide(s.collide_rect):
                    s.kill()
                    crashed.append(s)
            else:
                if spritecollide(s.rect):
                    s.kill()
                    crashed.append(s)
    else:
        for s in group:
            if s.collide_rect:
                if spritecollide(s.collide_rect):
                    crashed.append(s)
            else:
                if spritecollide(s.rect):
                    crashed.append(s)
    return crashed
